#
# compute scores with LSQ from paired comparisons
# and set origin from direct comparisons
#
# Author:  Christoph Dalitz
# Date:    26.10.2017
# License: GPL v2
#
#
# compare estimated scores from ML and LSQ with frequencies
#
# Author: Christoph Dalitz
#

library(minpack.lm)

# choice of distribution function
Fchoice <- "logis"

words <- read.csv("wordlist.txt", header=FALSE, stringsAsFactors=FALSE)
words <- words[,1]
n.words <- length(words)
pairs <- read.csv("results_paired.txt", sep=';', stringsAsFactors=FALSE)

# distribution function
sdnorm <- 1/3
scalelogis <- 1/pi/sqrt(3)
hunif <- 1/sqrt(3)
F.norm <- function(x) {
  return (pnorm(x, sd=sdnorm))
}
F.logis <- function(x) {
  return (plogis(x, scale=scalelogis))
}
F.unif <- function(x) {
  return (punif(x, min=-hunif, max=hunif))
}
Fprime.norm <- function(x) {
  return (dnorm(x, sd=sdnorm))
}
Fprime.logis <- function(x) {
  return (dlogis(x, scale=scalelogis))
}
Fprime.unif <- function(x) {
  return (dunif(x, min=-hunif, max=hunif))
}

# compute interval midpoints from cut-labels
midpoints <- function(x){
  lower <- as.numeric(gsub(",.*","",gsub("\\(|\\[|\\)|\\]","", x)))
  upper <- as.numeric(gsub(".*,","",gsub("\\(|\\[|\\)|\\]","", x)))
  return((lower+upper)/2)
}

# sum of squares function
# r = the n scores
# S = list of scores
# C = matrix with pair frequencies
SS.norm <- function(r, S, C, sd=sdnorm) {
  n <- length(r)
  dr <- sweep(matrix(r, nrow=n, ncol=n), MAR=2, r, FUN="-")
  SS <- S - rowSums(C * pnorm(dr, sd=sdnorm))
  return(SS)
}
SS.logis <- function(r, S, C, scale=scalelogis) {
  n <- length(r)
  dr <- sweep(matrix(r, nrow=n, ncol=n), MAR=2, r, FUN="-")
  SS <- S - rowSums(C * plogis(dr, scale=scalelogis))
  return(SS)
}
SS.unif <- function(r, S, C, h=hunif) {
  n <- length(r)
  dr <- sweep(matrix(r, nrow=n, ncol=n), MAR=2, r, FUN="-")
  SS <- S - rowSums(C * punif(dr, min=-h, max=h))
  return(SS)
}

# select distribution function
if (Fchoice == "norm") {
  F <- F.norm
  Fprime <- Fprime.norm
  SS <- SS.norm
} else if (Fchoice == "logis") {
  F <- F.logis
  Fprime <- Fprime.logis
  SS <- SS.logis
} else if (Fchoice == "unif") {
  F <- F.unif
  Fprime <- Fprime.unif
  SS <- SS.unif
}

# cumulate data
recompute <- TRUE
if (recompute) {
  x <- data.frame(word=words, S=rep(0,n.words), D=rep(0,n.words), n=rep(0,n.words), stringsAsFactors=FALSE)
  y <- pairs
  for (i in 1:n.words) {
    w <- as.character(words[i])
    S <- sum(y[y$Wort1 == w, "Sieg1"], na.rm=TRUE) +
      sum(y[y$Wort2 == w, "Sieg2"], na.rm=TRUE)
    D <- sum(y[y$Wort1 == w, "Unentschieden"], na.rm=TRUE) +
      sum(y[y$Wort2 == w, "Unentschieden"], na.rm=TRUE)
    n <- sum(y$Wort1 == w & !is.na(y$Sieg1), na.rm=TRUE) +
      sum(y$Wort2 == w & !is.na(y$Sieg1), na.rm=TRUE)
    x[i,] <- data.frame(word=w, S=S, D=D, n=n, stringsAsFactors=FALSE)
  }
  S <- x$S + 0.5*x$D
  C <- table(factor(pairs$Wort1, levels=words), factor(pairs$Wort2, levels=words))
  C <- C + t(C)
}

# find LSQ solution
par <- rep(0, n.words)
names(par) <- words
res1 <- nls.lm(par=par, fn=SS, S=S, C=C, control=list(maxit=10000))
mm.scores <- res1$par
names(mm.scores) <- words

#
# shift data to fit with direct comparisons
#

# compute scores from direct assigments
direct <- read.csv("results_direct.txt", sep=':', head=FALSE)
row.names(direct) <- direct[,1]
d.scores <- rowMeans(direct[as.character(words),-1])
names(d.scores) <- words
s <- data.frame(x=mm.scores, y=d.scores, row.names=words, stringsAsFactors=FALSE)

# error function for shift rho
SE <- function(rho, r, rprime) {
  diffs <- sign(rho+r) != sign(rprime)
  return (sum((rho+r[diffs])^2))
}
res <- optimize(f=SE, interval=c(-20,20), r=s$x, rprime=s$y)
cat(sprintf("best shift with SE: rho=%f (minval=%f)\n", res$minimum, res$objective))
mm.scores <- mm.scores + res$minimum

#
# estimate t
#
f <- rep(0, n.words)
names(f) <- words
for (i in 1:nrow(y)) {
  row <- y[i,]
  if (!is.na(row$Unentschieden)) {
    f[row$Wort1] <- f[row$Wort1] + Fprime(mm.scores[row$Wort1] - mm.scores[row$Wort2])
    f[row$Wort2] <- f[row$Wort2] + Fprime(mm.scores[row$Wort2] - mm.scores[row$Wort1])
  }
}
mm.scores["t"] <- sum(f*x$D/2) / sum(f^2)

cat("result ist stored in variable mm.scores\n")

#
# plot result
#

# bin data for plotting frequencies
x.raw <- data.frame(dr=(mm.scores[pairs$Wort1]-mm.scores[pairs$Wort2]), wins=pairs$Sieg1, draws=pairs$Unentschieden, losses=pairs$Sieg2)
xx <- x.raw
direction <- runif(nrow(x.raw)) >= 0.5
for (i in nrow(xx)) {
  if (direction[i] & xx[i,"dr"] < 0) {
    xx[i,"dr"] <- (-1)*xx[i,"dr"]
    xx[i,"wins"] <- x.raw[i,"losses"]
    xx[i,"losses"] <- x.raw[i,"wins"]
  } else
  if (!direction[i] & xx[i,"dr"] > 0) {
    xx[i,"dr"] <- (-1)*xx[i,"dr"]
    xx[i,"wins"] <- x.raw[i,"losses"]
    xx[i,"losses"] <- x.raw[i,"wins"]
  }
}
bins <- cut(xx$dr, breaks=150)
wins <- tapply(xx$wins, bins, mean)
draws <- tapply(xx$draws, bins, mean)
losses <- tapply(xx$losses, bins, mean)
dr <- midpoints(names(wins))

par(mfrow=c(1,2))

plot(dr, wins, main="P(win)")
lines(dr, F(dr - mm.scores["t"]), col="red")

plot(dr, draws, main="P(draw)")
lines(dr, F(dr + mm.scores["t"]) - F(dr - mm.scores["t"]), col="red")
